"""
    DigiCloud Compute Instance Service.
"""
from rich.prompt import Confirm

from digicloud import schemas
from .base import Lister, ShowOne, Command
from ..cli import parseractions
from ..error_handlers import CLIError
from ..utils import is_tty


class ListInstance(Lister):
    """List instances."""
    schema = schemas.InstanceList(many=True)

    def get_parser(self, prog_name):
        parser = super(ListInstance, self).get_parser(prog_name)
        parser.add_argument(
            '--simple',
            help='Advanced instances',
            default=None,
            action='store_true'
        )

        parser.add_argument(
            '--advanced',
            help='Simple instances',
            default=None,
            action='store_true'
        )

        parser.add_argument(
            '--id',
            nargs='+',
            help='list of instance IDs to filter, e.g --id 1 2 3 ',
            required=False
        )

        return parser

    def get_data(self, parsed_args):
        args = (parsed_args.simple, parsed_args.advanced)
        if all(args):
            raise CLIError(
                [dict(msg="You need to specify one of --simple or --advanced")])

        query_params = {}
        if parsed_args.simple:
            query_params['type'] = 'simple'
        elif parsed_args.advanced:
            query_params['type'] = 'advanced'

        data = self.app.session.get('/instances', params=query_params)
        return data


class CreateInstance(ShowOne):
    """Create instance."""
    schema = schemas.InstanceDetails()

    def get_parser(self, prog_name):
        parser = super(CreateInstance, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Instance name'
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help='Instance description'
        )
        parser.add_argument(
            '--instance-type',
            required=True,
            metavar='<instance_type>',
            help='InstanceType name or ID'
        )
        parser.add_argument(
            '--root-volume-size',
            required=False,
            type=int,
            metavar='<root_volume_size>',
            help=('Optionally you can set the size of root volume, otherwise root volume '
                  'will be created according to the instance type')
        )

        parser.add_argument(
            '--root-volume-type',
            required=False,
            metavar='<root_volume_type>',
            choices=['SSD', 'ULTRA_DISK'],
            type=lambda value: str(value).upper().replace("-", "_"),
            help=(
                'Optionally you can set the type of root volume, '
                'could be SSD or ULTRA_DISK case insensitive, the default is SSD'
            )
        )

        parser.add_argument(
            '--image',
            required=False,
            metavar='<image>',
            help='Image name or ID'
        )
        parser.add_argument(
            '--snapshot',
            required=False,
            metavar='<snapshot>',
            help='Snapshot name or ID'
        )
        parser.add_argument(
            '--network',
            required=False,
            metavar='<network>',
            help='Network name or ID'
        )
        parser.add_argument(
            '--ssh-key',
            metavar='<ssh_key>',
            help='SSH key name'
        )

        parser.add_argument(
            '--firewall',
            metavar='<firewall>',
            default=[],
            action='append',
            help='Firewall name or ID'
        )
        parser.add_argument(
            '--simple',
            help='Create a instance quickly',
            default=None,
            action='store_true'
        )
        parser.add_argument(
            '--with-public-ip',
            help='Create and associate a Public IP automatically',
            default=None,
            action='store_true'
        )

        parser.add_argument(
            '--public-ip',
            required=False,
            metavar='<public_ip>',
            help='Public IP name or ID to use after instance creation'
        )

        parser.add_argument(
            '--additional-volume',
            metavar='size=<size>,type=<type>',
            dest='additional_volumes',
            action=parseractions.MultiKeyValueAction,
            required_keys=['size', 'type'],
            default=[],
            help='Create and attach additional volume automatically, e.g --additional-volume size=500,type=SSD'
        )

        parser.add_argument(
            '--advanced',
            help='Create a instance with more configuration parameters',
            default=None,
            action='store_true'
        )

        parser.add_argument(
            '--count',
            metavar='<count>',
            type=int,
            help='Create multiple instance at once',
            default=1,
            required=False
        )

        return parser

    def _check_arg_validity(self, parsed_args):
        min_windows_root_vol_size = 20
        rules = [
            (
                all((parsed_args.advanced, parsed_args.simple)),
                "--advanced and --simple must not be used together",
            ),
            (
                parsed_args.simple and parsed_args.network is not None,
                "--simple and --network must not be used together",
            ),
            (
                parsed_args.simple and parsed_args.firewall,
                "--simple and --firewall must not be used together",
            ),
            (
                parsed_args.advanced and parsed_args.network is None,
                "--advanced requires --network to be present",
            ),
            (
                not parsed_args.simple and parsed_args.network is None,
                "--network is mandatory in advanced mode",
            ),
            (
                parsed_args.with_public_ip and parsed_args.public_ip is not None,
                "--with-public-ip and --public-ip must not be used together",
            ),
            # Creating multiple instance conditions:
            (
                parsed_args.with_public_ip and parsed_args.count > 1,
                "--with-public-ip and --count must not be used together",
            ),
            (
                parsed_args.public_ip and parsed_args.count > 1,
                "--public-ip and --count must not be used together",
            ),
            (
                any([
                    str(volume['type']).upper() not in ('SSD', 'ULTRA_DISK')
                    for volume in parsed_args.additional_volumes
                ]),
                "additional volume type must be either 'SSD' or 'ULTRA_DISK'"
            ),
            (
                parsed_args.image and parsed_args.snapshot,
                "--image and --snapshot must not be used together",
            ),
            (
                parsed_args.image is None and parsed_args.snapshot is None,
                "one of the --image and --snapshot arguments must be used",
            ),
            (
                parsed_args.count and parsed_args.count <= 0,
                "--count must be greater than 0",
            ),
            (
                parsed_args.image
                    and parsed_args.image.lower().startswith('windows')
                    and parsed_args.root_volume_size 
                    and parsed_args.root_volume_size is not None
                    and parsed_args.root_volume_size < min_windows_root_vol_size,
                "Minimum required root volume size for Windows images is 20 Gigabytes"
            ),

        ]
        errors = []
        for is_invalid, err_msg in rules:
            if is_invalid:
                errors.append(dict(msg=err_msg))
        if errors:
            raise CLIError(errors)

    def get_data(self, parsed_args):
        self._check_arg_validity(parsed_args)
        if parsed_args.count and parsed_args.count > 1:
            return self.create_instance_group(parsed_args)

        payload = {
            'name': parsed_args.name,
            'instance_type': parsed_args.instance_type,
            'type': 'simple' if parsed_args.simple else 'advanced',
        }
        if parsed_args.image:
            payload['image'] = parsed_args.image
        if parsed_args.snapshot:
            payload['snapshot'] = parsed_args.snapshot
        if parsed_args.network:
            payload['network'] = parsed_args.network
        if parsed_args.ssh_key:
            payload['ssh_key_name'] = parsed_args.ssh_key
        if parsed_args.firewall:
            payload['security_groups'] = parsed_args.firewall
        if parsed_args.additional_volumes:
            payload['additional_volumes'] = [
                {
                    "size": additional_volume['size'],
                    "volume_type": additional_volume['type'],
                } for additional_volume in parsed_args.additional_volumes
            ]
        if parsed_args.with_public_ip:
            payload['has_public_ip'] = parsed_args.with_public_ip
        if parsed_args.public_ip:
            payload['public_ip'] = parsed_args.public_ip
        if parsed_args.description:
            payload['description'] = parsed_args.description

        if parsed_args.root_volume_type:
            payload['root_volume_type'] = parsed_args.root_volume_type

        if parsed_args.root_volume_size:
            payload['root_volume_size'] = parsed_args.root_volume_size

        data = self.app.session.post('/instances', payload)
        return data

    def create_instance_group(self, parsed_args):
        self.schema = schemas.InstanceGroupDetails(many=False)
        payload = {
            'name': parsed_args.name,
            "count": parsed_args.count,
            "template": {
                'instance_type': parsed_args.instance_type,
                'type': 'simple' if parsed_args.simple else 'advanced',
            },
        }
        if parsed_args.image:
            payload['template']['image'] = parsed_args.image
        if parsed_args.snapshot:
            payload['template']['snapshot'] = parsed_args.snapshot
        if parsed_args.network:
            payload['template']['network'] = parsed_args.network
        if parsed_args.ssh_key:
            payload['template']['ssh_key_name'] = parsed_args.ssh_key
        if parsed_args.firewall:
            payload['template']['security_groups'] = parsed_args.firewall
        if parsed_args.description:
            payload['description'] = parsed_args.description
        if parsed_args.root_volume_type:
            payload['template']['root_volume_type'] = parsed_args.root_volume_type
        if parsed_args.root_volume_size:
            payload['template']['root_volume_size'] = parsed_args.root_volume_size
        if parsed_args.additional_volumes:
            payload['template']['additional_volumes'] = [
                {
                    "size": additional_volume['size'],
                    "volume_type": additional_volume['type'],
                } for additional_volume in parsed_args.additional_volumes
            ]

        data = self.app.session.post('/instance-groups', payload)
        command = "[bold blue]digicloud instance list --id %s[/bold blue]" % " ".join(
            data['instance_ids']
        )
        self.app.console.print(f"Check you new instances via `%s`" % command)
        return data


class ShowInstance(ShowOne):
    """Show instance details."""
    schema = schemas.InstanceDetails()

    def get_parser(self, prog_name):
        parser = super(ShowInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/instances/%s' % parsed_args.instance
        data = self.app.session.get(uri)
        return data


class DeleteInstance(Command):
    """Delete a instance."""

    def get_parser(self, prog_name):
        parser = super(DeleteInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )

        parser.add_argument(
            '--delete-volumes',
            help='Delete additional volumes attached to instance',
            default=None,
            action='store_true'
        )

        parser.add_argument(
            '--delete-root-volume',
            help='Use this switch to also delete root volume',
            default=None,
            action='store_true'
        )

        parser.add_argument(
            '--i-am-sure',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )

        return parser

    def take_action(self, parsed_args):
        snapshot_check_uri = '/instances/%s/snapshots' % parsed_args.instance
        instance_snapshot = self.app.session.get(snapshot_check_uri)
        if instance_snapshot:
            self.app.stdout.write(
                "NOTE: You have snapshot on this instance so you can't delete its volumes\n"
            )
            return

        if not self.confirm_instance_deletion(parsed_args):
            return
        uri = '/instances/%s' % parsed_args.instance

        additional_volumes_ids = self._get_additional_volumes_ids(parsed_args.instance)
        if self.confirm_root_volume_deletion(parsed_args):
            uri += self._format_query_param(uri, "delete_root_volume=true")
        if additional_volumes_ids and self.confirm_additional_volume_deletion(parsed_args):
            uri += self._format_query_param(uri, "delete_additional_volumes=true")

        self.app.session.delete(uri)

    def _has_query_param(self, uri: str):
        return '?' in uri

    def _format_query_param(self, uri, new_query_param: str):
        new_query_param = new_query_param.strip()
        return f"&{new_query_param}" if self._has_query_param(uri) else f"?{new_query_param}"


    def confirm_instance_deletion(self, parsed_args):
        if parsed_args.i_am_sure:
            return True
        if is_tty():
            instance = self.app.session.get('/instances/%s' % parsed_args.instance)
            user_response = Confirm.ask(
                "You're about to delete the instance named [red bold]{}[/red bold]. "
                "Are you sure?".format(
                    instance['name']
                ), default=False
            )
            if user_response:
                return True
            self.app.stdout.write("Operation cancelled by user\n")
        else:
            self.app.stderr.write(
                "Unable to perform 'instance delete' operation in non-interactive mode,"
                " without '--i-am-sure' switch\n")
            return False

    def confirm_root_volume_deletion(self, parsed_args):
        return self._volume_deletion_confirmation(
                parsed_args.delete_root_volume,
                "Would you like to also delete the root volume of this instance?"
        )

    def confirm_additional_volume_deletion(self, parsed_args):
        return self._volume_deletion_confirmation(
                parsed_args.delete_volumes,
                "Would you like to also delete the additional volumes attached to this instance?"
        )

    def _volume_deletion_confirmation(self, already_confirmed, msg):
        if already_confirmed:
            return True
        if is_tty():
            return Confirm.ask(
                msg,
                default=False
            )
        return False

    def _get_additional_volumes_ids(self, instance_name):
        uri = '/instances/%s/volumes' % instance_name
        non_root_volumes_id = [vol['id'] for vol in self.app.session.get(uri) if vol['is_root'] == False]
        return non_root_volumes_id

class ListInstanceVolume(Lister):
    """List instance volume(s)."""
    schema = schemas.InstanceVolume(many=True)

    def get_parser(self, prog_name):
        parser = super(ListInstanceVolume, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )

        return parser

    def get_data(self, parsed_args):
        # TODO: Could use a little bit of cache maybe
        # TODO: Duplicate logic
        volumes = {
            v['id']: v for v in self.app.session.get('/volumes')
        }
        instance_info = self.app.session.get('/instances/%s' % parsed_args.instance)
        uri = '/instances/%s/volumes' % instance_info['id']
        instance_volumes = self.app.session.get(uri)
        return [
            {
                "instance": instance_info['name'],
                "volume": volumes[volume['id']]['name'],
                **volume
            }
            for volume in instance_volumes

        ]


class ShowInstanceVolume(ShowOne):
    """Show instance volume details."""
    schema = schemas.InstanceVolume(many=False)

    def get_parser(self, prog_name):
        parser = super(ShowInstanceVolume, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        parser.add_argument(
            '--volume',
            required=True,
            metavar='<volume>',
            help='Volume name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        # TODO: Could use a little bit of cache maybe
        # TODO: Duplicate logic
        volumes = {
            v['id']: v for v in self.app.session.get('/volumes')
        }
        instance_info = self.app.session.get('/instances/%s' % parsed_args.instance)
        uri = '/instances/%s/volumes/%s' % (parsed_args.instance, parsed_args.volume)
        instance_volume = self.app.session.get(uri)
        return {
            "instance": instance_info['name'],
            "volume": volumes[instance_volume['id']]['name'],
            **instance_volume
        }


class AttachInstanceVolume(ShowOne):
    """Attach instance volume."""
    schema = schemas.InstanceVolume(many=False)

    def get_parser(self, prog_name):
        parser = super(AttachInstanceVolume, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--volume',
            required=True,
            metavar='<volume>',
            help='Volume name or ID'
        )

        return parser

    def get_data(self, parsed_args):
        uri = '/instances/%s/volumes' % parsed_args.instance
        payload = {'id': parsed_args.volume}

        data = self.app.session.post(uri, payload)

        return data


class DetachInstanceVolume(Command):
    """Detach instance volume."""

    def get_parser(self, prog_name):
        parser = super(DetachInstanceVolume, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--volume',
            required=True,
            metavar='<volume>',
            help='Volume name or ID'
        )

        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/volumes/%s' % (parsed_args.instance, parsed_args.volume)
        self.app.session.delete(uri)


class ListInstanceInterface(Lister):
    """List instance interface(s)."""
    schema = schemas.ListInstanceInterface(many=True)

    def get_parser(self, prog_name):
        parser = super(ListInstanceInterface, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )

        return parser

    def get_data(self, parsed_args):
        uri = '/instances/%s/interfaces' % parsed_args.instance
        data = self.app.session.get(uri)

        return data


class AttachInterface(ShowOne):
    """Attach interface."""
    schema = schemas.InstanceInterface(many=False)

    def get_parser(self, prog_name):
        parser = super(AttachInterface, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--network',
            required=True,
            metavar='<network>',
            help='Network name or ID',

        )
        return parser

    def get_data(self, parsed_args):
        payload = {'net': parsed_args.network}
        uri = '/instances/%s/interfaces' % parsed_args.instance
        data = self.app.session.post(uri, payload)
        return data


class DetachInterface(Command):
    """Detach interface."""

    def get_parser(self, prog_name):
        parser = super(DetachInterface, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--interface-id',
            required=True,
            metavar='<interface>',
            help='Interface ID',

        )

        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/interfaces/%s' % (
            parsed_args.instance, parsed_args.interface_id)
        self.app.session.delete(uri)


class UpdateInstance(ShowOne):
    """Update instance name, description or even resize it to another instance-type"""
    schema = schemas.InstanceDetails()

    def get_parser(self, prog_name):
        parser = super(UpdateInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance ID or name',
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            required=False,
            help='Instance new name, must be unique',
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            required=False,
            help='Instance description',
        )
        parser.add_argument(
            '--instance-type',
            metavar='<instance_type>',
            required=False,
            help='New instance type you want to use for this instance',
        )

        return parser

    def get_data(self, parsed_args):
        uri = '/instances/%s' % parsed_args.instance
        payload = {}
        if parsed_args.name:
            payload['name'] = parsed_args.name
        if parsed_args.description:
            payload['description'] = parsed_args.description
        if parsed_args.instance_type:
            payload['instance_type'] = parsed_args.instance_type
            self.app.console.print(
                "Resizing your instance might cause additional charges",
                style='bold yellow'
            )
        if not payload:
            raise CLIError([
                dict(
                    msg="At least one of --name or --description or "
                        "--instance-type must be provided"
                )
            ]
            )
        data = self.app.session.patch(uri, payload)

        return data

    def _on_400(self, parsed_args, response):
        error_msg = response.json()['message']
        if 'stop your instance' in error_msg:
            return CLIError([
                dict(
                    msg=error_msg,
                    hint="You can stop your instance by running: "
                         "[bold blue]digicloud instance stop {}[/bold blue]".format(
                        parsed_args.instance
                    )
                )
            ])

    def _on_404(self, parsed_args, response):
        return CLIError([
            dict(
                msg="Please check your instance name and try again!",
                hint="You can list your instances by running: "
                     "[bold blue]digicloud instance list [/bold blue]"
            )
        ])


class ResizeInstance(ShowOne):
    """Resize your instance"""
    schema = schemas.InstanceDetails()

    def get_parser(self, prog_name):
        parser = super(ResizeInstance, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance ID or name',
        )
        parser.add_argument(
            '--instance-type',
            metavar='<instance_type>',
            required=True,
            help='The instance type you want to use for this instance',
        )

        return parser

    def get_data(self, parsed_args):
        uri = '/instances/%s' % parsed_args.instance
        payload = {}
        if parsed_args.instance_type:
            payload['instance_type'] = parsed_args.instance_type
            self.app.console.print(
                "Resizing your instance might cause additional charges",
                style='bold yellow'
            )
        return self.app.session.patch(uri, payload)

    def _on_400(self, parsed_args, response):
        error_msg = response.json()['message']
        if 'stop your instance' in error_msg:
            return CLIError([
                dict(
                    msg=error_msg,
                    hint="You can stop your instance by running: "
                         "[bold blue]digicloud instance stop {}[/bold blue]".format(
                        parsed_args.instance
                    )
                )
            ])


class ListFirewall(Lister):
    """List Instance Firewalls"""
    schema = schemas.FirewallList(many=True)

    def get_parser(self, prog_name):
        parser = super(ListFirewall, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/instances/%s/security-groups' % parsed_args.instance
        data = self.app.session.get(uri)
        return data


class AddFirewall(Command):
    """Add Firewall to instance"""

    def get_parser(self, prog_name):
        parser = super(AddFirewall, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--firewall',
            required=True,
            metavar='<firewall>',
            help='Firewall name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        payload = {'security_group_ref': parsed_args.firewall}
        uri = '/instances/%s/security-groups' % parsed_args.instance
        self.app.session.post(uri, payload)


class RemoveFirewall(Command):
    """Remove Firewall from Instance"""

    def get_parser(self, prog_name):
        parser = super(RemoveFirewall, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        parser.add_argument(
            '--firewall',
            required=True,
            metavar='<firewall>',
            help='Firewall name or ID',
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/security-groups/%s' % (parsed_args.instance,
                                                    parsed_args.firewall)
        self.app.session.delete(uri)


class ListInstanceSnapshot(Lister):
    """List Instance Snapshots"""
    schema = schemas.SnapshotList(many=True)

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/instances/%s/snapshots' % parsed_args.instance
        data = self.app.session.get(uri)
        return data


class InstanceConsoleLog(Command):
    """Show instance console log."""

    def get_parser(self, prog_name):
        parser = super(InstanceConsoleLog, self).get_parser(prog_name)
        parser.add_argument(
            'instance',
            metavar='<instance>',
            help='Instance name or ID',
        )
        parser.add_argument(
            '--lines',
            help='Number of instance console log lines',
            required=False
        )
        parser.add_argument(
            '--full',
            help='Full of instance console log lines',
            default=None,
            action='store_true'
        )
        return parser

    def _check_arg_validity(self, parsed_args):
        rules = [
            (
                all((parsed_args.full, parsed_args.lines)),
                "--lines and --full must not be used together",
            )
        ]
        errors = []
        for is_invalid, err_msg in rules:
            if is_invalid:
                errors.append(dict(msg=err_msg))
        if errors:
            raise CLIError(errors)

    def take_action(self, parsed_args):
        default_lines = 100
        self._check_arg_validity(parsed_args)
        uri = '/instances/%s/logs' % parsed_args.instance

        query_params = {}
        if parsed_args.full:
            query_params['lines'] = None
        elif parsed_args.lines:
            query_params['lines'] = parsed_args.lines
        else:
            query_params['lines'] = default_lines

        data = self.app.session.get(uri, params=query_params)
        self.app.console.print(data)


"""
    Common Utilitie functions.
"""
import re
import sys
import uuid

import requests


def tabulate(data):
    """Make dict or list to table ready tuples."""
    if isinstance(data, dict) and data:
        headers = data.keys()
        rows = [value for key, value in data.items()]
    elif isinstance(data, list) and data:
        headers = data[0].keys()
        rows = [[value for key, value in obj.items()] for obj in data]
    else:
        return (), ()

    return headers, rows


def is_tty():
    return sys.stdin.isatty()


def get_latest_version_number(package_name):
    uri = 'https://pypi.org/pypi/{}/json'.format(package_name)
    return requests.get(uri, timeout=1).json()['info']['version']


def compare_versions(left, right):
    v1_parts = list(map(int, left.split('.')))
    while len(v1_parts) < 3:
        v1_parts.append(0)

    v2_parts = list(map(int, right.split('.')))
    while len(v2_parts) < 3:
        v2_parts.append(0)
    return {
        'major': v1_parts[0] - v2_parts[0],
        'minor': v1_parts[1] - v2_parts[1],
        'patch': v1_parts[2] - v2_parts[2],
    }


def convert_camel_to_snake(name):
    return re.sub(r'(?<!^)(?=[A-Z])', '_', name).lower()


def t_or_f_or_n(arg):
    upper_arg = str(arg).upper()
    if 'TRUE'.startswith(upper_arg):
        return True
    elif 'FALSE'.startswith(upper_arg):
        return False
    else:
        return None



"""
    DigiCloud Security Service.
"""

from .base import Lister, ShowOne, Command
from .. import schemas


class ListFirewall(Lister):
    """List Firewalls."""
    schema = schemas.FirewallList(many=True)

    def get_data(self, parsed_args):
        data = self.app.session.get('/security-groups')

        return data


class ShowFirewall(ShowOne):
    """Show Firewall details."""
    schema = schemas.FirewallDetails()

    def get_parser(self, prog_name):
        parser = super(ShowFirewall, self).get_parser(prog_name)
        parser.add_argument(
            'firewall',
            metavar='<firewall>',
            help='firewall name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/security-groups/%s' % parsed_args.firewall
        data = self.app.session.get(uri)

        return data


class CreateFirewall(ShowOne):
    """Create Firewall."""
    schema = schemas.FirewallDetails()

    def get_parser(self, prog_name):
        parser = super(CreateFirewall, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Firewall name'
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            default="",
            required=False,
            help='Firewall description'
        )

        return parser

    def get_data(self, parsed_args):
        payload = {'name': parsed_args.name, 'description': parsed_args.description}
        data = self.app.session.post('/security-groups', payload)

        return data


class DeleteFirewall(Command):
    """Delete Firewall."""

    def get_parser(self, prog_name):
        parser = super(DeleteFirewall, self).get_parser(prog_name)
        parser.add_argument(
            'firewall',
            metavar='<firewall>',
            help='Firewall name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/security-groups/%s' % parsed_args.firewall
        self.app.session.delete(uri)


class UpdateFirewall(Command):
    """Update Firewall."""

    def get_parser(self, prog_name):
        parser = super(UpdateFirewall, self).get_parser(prog_name)
        parser.add_argument(
            'firewall',
            metavar='<firewall>',
            help='Firewall name or ID'
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            required=False,
            help='Firewall name'
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            required=False,
            help='Firewall description'
        )
        return parser

    def take_action(self, parsed_args):
        uri = f'/security-groups/{parsed_args.firewall}'
        payload = {}
        if parsed_args.name:
            payload['name'] = parsed_args.name
        if parsed_args.description:
            payload['description'] = parsed_args.description
        self.app.session.patch(uri, payload)


class ListFirewallRule(Lister):
    """List Firewall Rules."""
    schema = schemas.FirewallRuleList(many=True)

    def get_parser(self, prog_name):
        parser = super(ListFirewallRule, self).get_parser(prog_name)
        parser.add_argument(
            'firewall',
            metavar='<firewall>',
            help='Firewall rule ID',
        )
        return parser

    def get_data(self, parsed_args):
        data = self.app.session.get(
            '/security-groups/{}/rules'.format(parsed_args.firewall)
        )
        return data


class ShowFirewallRule(ShowOne):
    """Show Firewall rule details."""
    schema = schemas.FirewallRuleDetails()

    def get_parser(self, prog_name):
        parser = super(ShowFirewallRule, self).get_parser(prog_name)
        parser.add_argument(
            'firewall_rule',
            metavar='<firewall_rule>',
            help='Firewall rule ID',
        )
        parser.add_argument(
            '--firewall',
            required=True,
            metavar='<firewall>',
            help='Firewall name or ID'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/security-groups/{}/rules/{}'.format(
            parsed_args.firewall,
            parsed_args.firewall_rule
        )
        data = self.app.session.get(uri)
        return data


class CreateFirewallRule(ShowOne):
    """Create Firewall Rule."""
    schema = schemas.FirewallRuleDetails()

    def get_parser(self, prog_name):
        parser = super(CreateFirewallRule, self).get_parser(prog_name)
        parser.add_argument(
            'firewall',
            metavar='<firewall>',
            help='Firewall name or ID'
        )
        parser.add_argument(
            '--direction',
            required=True,
            metavar='<direction>',
            help='Direction'
        )
        parser.add_argument(
            '--ethertype',
            default='IPv4',
            metavar='<ethertype>',
            help='IP version',
            choices=['IPv4', 'IPv6']
        )
        parser.add_argument(
            '--port-range-min',
            metavar='<port_range_min>',
            help='Minimum of port range'
        )
        parser.add_argument(
            '--port-range-max',
            metavar='<port_range_max>',
            help='Maximum of port range'
        )
        parser.add_argument(
            '--protocol',
            required=True,
            metavar='<protocol>',
            help="Protocol"
        )
        parser.add_argument(
            '--remote-ip-prefix',
            required=False,
            metavar='<remote_ip_prefix>',
            help='remote IP prefix matched by this Firewall rule'
        )
        return parser

    def get_data(self, parsed_args):
        args = [
            'direction', 'ethertype', 'port_range_min',
            'port_range_max', 'protocol', 'remote_ip_prefix',
        ]
        payload = {arg: getattr(parsed_args, arg) for arg in args
                   if getattr(parsed_args, arg) is not None}
        uri = '/security-groups/%s/rules' % parsed_args.firewall
        data = self.app.session.post(uri, payload)

        return data


class DeleteFirewallRule(Command):
    """Delete Firewall Rule."""

    def get_parser(self, prog_name):
        parser = super(DeleteFirewallRule, self).get_parser(prog_name)
        parser.add_argument(
            'firewall_rule',
            metavar='<firewall_rule>',
            help='Firewall rule ID'
        )
        parser.add_argument(
            '--firewall',
            required=True,
            metavar='<firewall>',
            help='Firewall name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/security-groups/{}/rules/{}'.format(parsed_args.firewall,
                                                    parsed_args.firewall_rule)
        self.app.session.delete(uri)



import os


class DocumentationContainer:
    def __init__(self, docs_dir):
        self.docs = self._discover_docs(docs_dir)

    def _discover_docs(self, docs_dir):
        docs = {}
        for file_name in os.listdir(docs_dir):
            if file_name.endswith('md'):
                docs[file_name.split('.')[0]] = os.path.join(docs_dir, file_name)
        return docs

    def get(self, topic):
        doc_file = self.docs.get(topic.lower())
        if doc_file is None:
            raise FileNotFoundError("No topic available named {}".format(topic))
        with open(doc_file) as doc_file_ptr:
            return doc_file_ptr.read()

    def list(self):
        return sorted(self.docs.keys())



from datetime import datetime

import tzlocal
import pytz
from marshmallow import Schema, fields, INCLUDE, post_dump

from digicloud import formatter


class APIDatetime(fields.Str):
    default_input_format = None  # Default format is  ISO_8601
    default_tz = pytz.UTC  # API uses UTC
    default_output_format = "%Y/%m/%d %H:%M"

    def _serialize(self, value, attr, obj, **kwargs):
        try:
            input_tz_datetime = self._parse_datetime_str(value)
            output_tz_datetime = self._get_user_tz_datetime(input_tz_datetime)
            return output_tz_datetime.strftime(self.default_output_format)
        except Exception:

            return "N/A"

    def _get_user_tz_datetime(self, dt_value):
        if self.metadata.get("as_timezone"):
            tz = pytz.timezone(self.metadata.get("as_timezone"))
        else:
            tz = tzlocal.get_localzone()
        return dt_value.astimezone(tz)

    def _parse_datetime_str(self, datetime_str):
        fmt = self.metadata.get("datetime_format", self.default_input_format)
        if fmt is None:
            naive_dt = datetime.fromisoformat(datetime_str)
        else:
            naive_dt = datetime.strptime(datetime_str, fmt)
        return naive_dt.replace(tzinfo=self.default_tz)  # API is always UTC


class NetworkList(Schema):
    name = fields.Str()
    description = fields.Str()
    subnets = fields.Function(lambda s: ", ".join([
        subnet['name'] for subnet in s.get('subnets', [])
    ]) or "N/A")
    has_gateway_interface = fields.Boolean()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class NetworkDetail(Schema):
    id = fields.Str()
    name = fields.Str()
    description = fields.Str()
    status = fields.Str()
    admin_state = fields.Str()
    subnets = fields.List(fields.Dict(), )
    mtu = fields.Int()
    updated_at = APIDatetime()
    created_at = APIDatetime()

    class Meta:
        ordered = True


class PublicIPList(Schema):
    name = fields.Str()
    public_ip_address = fields.Str()
    description = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class PublicIPDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    description = fields.Str()
    public_ip_address = fields.Str()
    status = fields.Str()
    router_id = fields.Str()
    updated_at = APIDatetime()
    created_at = APIDatetime()

    class Meta:
        ordered = True


class RouterList(Schema):
    name = fields.Str()
    description = fields.Str()
    has_gateway = fields.Str()
    gateway_address = fields.Str()
    admin_state = fields.Str()
    status = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class RouterDetails(RouterList):
    id = fields.Str()
    routes = fields.List(fields.Dict)

    class Meta:
        ordered = True


class RouterInterfaceList(Schema):
    id = fields.Str()
    network = fields.Str(attribute='network_name')
    subnet = fields.Str(attribute='subnet_name')
    ip_address = fields.Str()
    status = fields.Str()
    admin_state = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class RouterInterfaceDetails(RouterInterfaceList):
    firewalls = fields.List(fields.Str, attribute='security_groups')
    created_at = APIDatetime()

    class Meta:
        ordered = True


class FirewallRuleList(Schema):
    id = fields.Str()
    ethertype = fields.Str()
    direction = fields.Str()
    protocol = fields.Str()
    port_range_max = fields.Int()
    port_range_min = fields.Int()
    remote_ip_prefix = fields.Str()
    description = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class FirewallRuleDetails(Schema):
    id = fields.Str()
    firewall_id = fields.Str(attribute='security_group_id')
    ethertype = fields.Str()
    direction = fields.Str()
    protocol = fields.Str()
    port_range_max = fields.Int()
    port_range_min = fields.Int()
    remote_ip_prefix = fields.Str()
    description = fields.Str()
    updated_at = APIDatetime()
    created_at = APIDatetime()

    class Meta:
        ordered = True


class FirewallList(Schema):
    name = fields.Str()
    description = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class FirewallDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    description = fields.Str()
    updated_at = APIDatetime()
    created_at = APIDatetime()

    class Meta:
        ordered = True


class ImageList(Schema):
    name = fields.Str()

    class Meta:
        ordered = True


class SSHKeyList(Schema):
    name = fields.Str()
    finger_print = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class SSHKeyDetails(SSHKeyList):
    id = fields.Str()
    updated_at = APIDatetime()
    created_at = APIDatetime()


class ImageDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    size = fields.Function(lambda v: formatter.format_size(v['size'], multiplier=1))
    default_user = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class InstanceList(Schema):
    name = fields.Str()
    description = fields.Str()
    type = fields.Str()
    addresses = fields.List(fields.Str())
    instance_type = fields.Str()
    status = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class InstanceDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    key_name = fields.Str()
    firewalls = fields.List(fields.Str(), attribute='security_groups')
    addresses = fields.List(fields.Str())
    access_ip = fields.Str()
    cores = fields.Str()
    memory = fields.Function(lambda v: formatter.format_ram(int(v['memory'])))
    root_volume_size = fields.Integer()
    root_volume_type = fields.Str()
    instance_type = fields.Str()
    status = fields.Str()
    type = fields.Str()
    os_name = fields.Str()
    os_username = fields.Str()
    namespace_id = fields.Str()
    description = fields.Str()
    created_by = fields.Str()
    updated_at = APIDatetime()
    created_at = APIDatetime()

    class Meta:
        ordered = True


class InstanceGroupDetails(Schema):
    name = fields.Str()
    instance_ids = fields.List(fields.Str())

    class Meta:
        ordered = True


class InstanceVolume(Schema):
    volume = fields.Str()
    instance = fields.Str()
    volume_type = fields.Str()
    device = fields.Str()
    size = fields.Function(lambda v: formatter.format_size(v['size']))

    class Meta:
        ordered = True


class InstanceInterface(Schema):
    id = fields.Str()
    net_id = fields.Str()
    mac_addr = fields.Str()
    port_state = fields.Str()
    fixed_ips = fields.List(fields.Dict())

    class Meta:
        ordered = True


class ListInstanceInterface(Schema):
    id = fields.Str()
    network_id = fields.Str()
    mac_address = fields.Str()
    port_state = fields.Str()
    public_ip = fields.Str()
    port_security_enabled = fields.Boolean()
    fixed_ips = fields.List(fields.Dict())

    class Meta:
        ordered = True


class InstanceInterfaceDetail(Schema):
    id = fields.Str()
    admin_state = fields.Str()
    mac_addr = fields.Str()
    port_security_enabled = fields.Bool()
    updated_at = APIDatetime(datetime_format="%Y-%m-%dT%H:%M:%SZ")
    created_at = APIDatetime(datetime_format="%Y-%m-%dT%H:%M:%SZ")

    class Meta:
        ordered = True


class SnapshotList(Schema):
    name = fields.Str()
    description = fields.Str()
    status = fields.Str()
    updated_at = fields.Str()

    class Meta:
        ordered = True


class SnapshotDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    description = fields.Str()
    status = fields.Str()
    volume_ids = fields.List(fields.Str)
    created_at = fields.Str()
    updated_at = fields.Str()

    class Meta:
        ordered = True


class InstanceType(Schema):
    name = fields.Str()
    family = fields.Str()
    ram = fields.Function(lambda v: formatter.format_ram(v['ram']),
                          data_key="Memory (GB)", )
    disk = fields.Int(data_key="Disk (GB)")
    vcpus = fields.Str()

    class Meta:
        ordered = True


class RouterStaticRouteSchema(Schema):
    destination = fields.Str()
    nexthop = fields.Str()


class IPAllocationPool(Schema):
    start = fields.Str(required=True)
    end = fields.Str(required=True)


class SubnetList(Schema):
    name = fields.Str()
    description = fields.Str()
    cidr = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class SubnetDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    description = fields.Str()
    ip_version = fields.Str()
    network_id = fields.Str()
    gateway_ip = fields.Str()
    cidr = fields.Str()
    enable_dhcp = fields.Boolean()
    allocation_pools = fields.List(fields.Nested(IPAllocationPool))
    dns_nameservers = fields.List(fields.Str)
    host_routes = fields.List(fields.Nested(RouterStaticRouteSchema))
    updated_at = APIDatetime()
    created_at = APIDatetime()

    class Meta:
        ordered = True


class VolumeList(Schema):
    name = fields.Str()
    description = fields.Str()
    size = fields.Function(lambda v: formatter.format_size(v['size']))
    volume_type = fields.Str()
    status = fields.Str()
    attached_to = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class VolumeDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    description = fields.Str()
    size = fields.Function(lambda v: formatter.format_size(v['size']))
    volume_type = fields.Str()
    status = fields.Str()
    bootable = fields.Bool()
    instance = fields.Str(attribute='attachment_info.instance')
    device = fields.Str(attribute='attachment_info.device')
    attached_at = APIDatetime(
        attribute='attachment_info.attached_at',
        datetime_format="%Y-%m-%dT%H:%M:%S.%f")
    updated_at = APIDatetime()
    created_at = APIDatetime()

    class Meta:
        ordered = True


class NamespaceList(Schema):
    id = fields.Str()
    name = fields.Str()
    description = fields.Str()
    updated_at = APIDatetime(datetime_format="%Y-%m-%dT%H:%M:%S")

    class Meta:
        ordered = True


class NamespaceDetails(NamespaceList):
    updated_at = APIDatetime(datetime_format="%Y-%m-%dT%H:%M:%S")
    created_at = APIDatetime(datetime_format="%Y-%m-%dT%H:%M:%S")

    class Meta:
        ordered = True


class NamespaceMemberList(Schema):
    id = fields.Str()
    email = fields.Function(lambda data: data['email'] if 'email' in data else data['invitee_email'])
    role = fields.Str(default="")
    membership = fields.Method("_determine_membership")

    def _determine_membership(self, data):
        if 'invitation_date' in data:
            invitation_date = APIDatetime(datetime_format="%Y-%m-%dT%H:%M:%S").serialize('invitation_date', data)
            return f"invitation pending since {invitation_date}"
        else:
            return "member"

    class Meta:
        ordered = True


class UserInvitation(Schema):
    id = fields.Str()
    namespace = fields.Function(lambda invite: invite['namespace']['name'])
    invited_by = fields.Function(lambda invite: "{} {} ({})".format(
        invite['inviter']['first_name'],
        invite['inviter']['last_name'],
        invite['inviter']['email'],
    ))
    invitation_date = APIDatetime(datetime_format="%Y-%m-%dT%H:%M:%S")

    class Meta:
        ordered = True


class NamespaceQuota(Schema):
    id = fields.Str()
    description = fields.Str()
    quota = fields.Int()
    used = fields.Int(allow_none=True)

    class Meta:
        ordered = True


class ErrorSchema(Schema):
    error_code = fields.Str()
    message = fields.Str()

    class Meta:
        unknown = INCLUDE


class ExternalVpnConnectionDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    description = fields.Str()
    auth_mode = fields.Str()
    psk = fields.Str()
    initiator = fields.Str()
    admin_state_up = fields.Bool()
    mtu = fields.Int()
    dpd = fields.Dict()
    peer_address = fields.Str()
    peer_id = fields.Str()
    # IKE policy attrs
    ike_auth_algorithm = fields.Function(lambda data: data["ike_policy"]["auth_algorithm"])
    ike_encryption_algorithm = fields.Function(
        lambda data: data["ike_policy"]["encryption_algorithm"])
    ike_pfs = fields.Function(lambda data: data["ike_policy"]["pfs"])
    ike_lifetime = fields.Function(lambda data: data["ike_policy"]["lifetime"])
    ike_version = fields.Function(lambda data: data["ike_policy"]["ike_version"])
    # VPN service attrs
    router_id = fields.Function(lambda data: data["vpn_service"]["router_id"])
    # IpSec policy attrs
    ipsec_auth_algorithm = fields.Function(lambda data: data["ipsec_policy"]["auth_algorithm"])
    ipsec_encapsulation_mode = fields.Function(
        lambda data: data["ipsec_policy"]["encapsulation_mode"])
    ipsec_encryption_algorithm = fields.Function(
        lambda data: data["ipsec_policy"]["encryption_algorithm"])
    ipsec_pfs = fields.Function(lambda data: data["ipsec_policy"]["pfs"])
    ipsec_transform_protocol = fields.Function(
        lambda data: data["ipsec_policy"]["transform_protocol"])
    ipsec_lifetime = fields.Function(
        lambda data: data["ipsec_policy"]["lifetime"]["value"])
    # endpoint group attrs
    local_endpoint_group = fields.Function(
        lambda data: data["local_endpoint_group"]["endpoints"])
    peer_endpoint_group = fields.Function(
        lambda data: data["peer_endpoint_group"]["endpoints"])
    status = fields.Str()
    created_at = APIDatetime()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class ExternalVpnConnectionInList(Schema):
    name = fields.Str()
    description = fields.Str()
    initiator = fields.Str()
    admin_state_up = fields.Bool()
    mtu = fields.Int()
    auth_mode = fields.Str()
    peer_address = fields.Str()
    status = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class QuotaRequest(Schema):
    quota_id = fields.Str()
    required_quota = fields.Int()
    status = fields.Str()
    updated_at = fields.Str()

    class Meta:
        ordered = True


class EdgeDomainList(Schema):
    id = fields.Str()
    name = fields.Str()
    status = fields.Str()
    record_count = fields.Integer()
    is_subdomain = fields.Boolean()
    ns_verification = fields.Str()

    class Meta:
        ordered = True


class EdgeDomainDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    is_subdomain = fields.Boolean()
    ns_verification = fields.Str()

    class Meta:
        ordered = True


class EdgeDomainPackageDetails(Schema):
    id = fields.Str()
    package_name_en = fields.Str()
    package_name_fa = fields.Str()

    class Meta:
        ordered = True


class DomainDNSSECDetails(Schema):
    dnssec = fields.Boolean()
    ds = fields.Str()

    class Meta:
        ordered = True


class RecordListSchema(Schema):
    available_keys = ("id", "name", "ttl", "note", "type", "ip_address", "proxy", "content", "port", "weight", "proto",
                      "service", "target", "priority", "mail_server",)
    id = fields.Str()
    name = fields.Str(required=True)
    type = fields.Str(required=True)
    ttl = fields.Str(required=True)
    note = fields.Str(required=False)
    proxy = fields.Boolean(required=False)

    class Meta:
        ordered = True


class ARecordDetailsSchema(RecordListSchema):
    ip_address = fields.Str(required=True)
    proxy = fields.Boolean(required=False, missing=False)


class TXTRecordDetailsSchema(RecordListSchema):
    content = fields.Str(required=True)


class CNAMERecordDetailsSchema(RecordListSchema):
    target = fields.Str(required=True)
    proxy = fields.Boolean(required=False, missing=False)


class MXRecordDetailsSchema(RecordListSchema):
    mail_server = fields.Str(required=True)
    priority = fields.Integer(required=True)


class SRVRecordDetailsSchema(RecordListSchema):
    port = fields.Integer(required=True)
    weight = fields.Integer(required=True)
    proto = fields.Str(required=True)
    service = fields.Str(required=True)
    target = fields.Str(required=True)
    priority = fields.Integer(required=True)


class HealthCheckDetail(Schema):
    id = fields.Str()
    name = fields.Str()
    openstack_id = fields.Str()
    delay = fields.Int()
    max_retries = fields.Int()
    max_retries_down = fields.Int()
    timeout = fields.Int()
    type = fields.Str()
    operating_status = fields.Str()
    created_at = APIDatetime()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class BackendMemberDetail(Schema):
    id = fields.Str()
    ip_address = fields.Str()
    port = fields.Str()
    operating_status = fields.Str()
    created_at = APIDatetime()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class BackendDetail(Schema):
    id = fields.Str()
    algorithm = fields.Str()
    protocol = fields.Str()
    health_checks = fields.Function(lambda lb: [hc["id"] for hc in lb["health_checks"]])
    members = fields.Function(lambda lb: [member["id"] for member in lb["members"]])
    updated_at = APIDatetime()
    created_at = APIDatetime()

    class Meta:
        ordered = True


class FrontendDetail(Schema):
    id = fields.Str()
    port = fields.Str()
    protocol = fields.Str()
    timeout_client_data = fields.Int()
    timeout_tcp_inspect = fields.Int()
    timeout_member_connect = fields.Int()
    timeout_member_data = fields.Int()
    connection_limit = fields.Int()
    updated_at = APIDatetime()
    created_at = APIDatetime()
    default_backend_id = fields.Str()

    class Meta:
        ordered = True


class LoadBalancerList(Schema):
    name = fields.Str()
    description = fields.Str()
    ip_address = fields.Str()
    status = fields.Str()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class LoadBalancerDetail(Schema):
    id = fields.Str()
    name = fields.Str()
    description = fields.Str()
    status = fields.Str()
    backends = fields.Function(lambda lb: [backend["id"] for backend in lb["backends"]])
    frontends = fields.Function(lambda lb: [frontend["id"] for frontend in lb["frontends"]])
    ip_address = fields.Str()
    created_at = APIDatetime()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class AccessKeyDetails(Schema):
    access_key = fields.Str()
    secret_key = fields.Str()
    is_active = fields.Boolean()

    class Meta:
        ordered = True


class UpstreamDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    lb_method = fields.Str()
    keep_alive = fields.Int()
    ssl_policy = fields.Str()
    servers_count = fields.Int()

    class Meta:
        ordered = True


class UpstreamServerDetails(Schema):
    id = fields.Str()
    ip_domain = fields.Str()
    port = fields.Int()
    weight = fields.Int()
    fail_timeout = fields.Int()

    class Meta:
        ordered = True


class SSLDetails(Schema):
    id = fields.Str()
    type = fields.Str()
    policy = fields.Str()
    enable = fields.Boolean()
    hsts = fields.Boolean()
    https_redirect = fields.Boolean()
    ocsp_check = fields.Boolean()
    min_tls_versions = fields.Str()
    expires_at = APIDatetime()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class LocationDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    path = fields.Str()
    path_case_sensitive = fields.Bool()
    upstream_id = fields.Str()
    origin_headers = fields.List(fields.Dict())
    response_headers = fields.List(fields.Dict())
    updated_at = APIDatetime()
    cache_enabled = fields.Bool()
    cache_ttl = fields.Str()
    cache_key = fields.Str()
    cache_cookie_name = fields.Str()
    cache_zone = fields.Str()
    secure_link_enable = fields.Bool()
    secure_link_secret = fields.Str()
    redirect_enable = fields.Bool()
    redirect_url = fields.Str()
    redirect_status_code = fields.Integer()
    rate_limit_id = fields.Str()

    @post_dump
    def convert_redirect_status_code(self, data, **kwargs):
        redirect_status_code_dict = {
            301: "permanent",
            302: "temporary",
        }
        data["redirect_status_code"] = redirect_status_code_dict[
            data["redirect_status_code"]
        ]

        if not data['redirect_enable']:
            data['redirect_status_code'] = None

        return data

    class Meta:
        ordered = True


class EdgeFirewallDetails(Schema):
    id = fields.Str()
    location_id = fields.Str()
    input = fields.Str()
    value = fields.Str()
    action = fields.Str()
    operator = fields.Str()
    priority = fields.Int()
    updated_at = APIDatetime()

    class Meta:
        ordered = True


class CacheZoneList(Schema):
    id = fields.Str()
    name = fields.Str()

    class Meta:
        ordered = True


class GeneralSettingSchema(Schema):
    id = fields.Str()
    origin_headers = fields.List(fields.Dict())
    response_headers = fields.List(fields.Dict())
    developer_mode = fields.Boolean()
    maintenance_mode = fields.Boolean()
    redirect_to_www = fields.Boolean()
    ip_geolocation = fields.Boolean()
    intercept_errors = fields.Boolean()
    max_upload_size = fields.Int()
    custom_host_header = fields.Str()
    created_at = fields.Str(dump_only=True)
    updated_at = fields.Str(dump_only=True)

    class Meta:
        ordered = True


class RateLimitDetails(Schema):
    id = fields.Str()
    name = fields.Str()
    time = fields.Str()
    requests = fields.Int()
    burst = fields.Int()
    exclusion_list = fields.List(fields.Str())
    created_at = fields.Str()
    updated_at = fields.Str()

    class Meta:
        ordered = True


def handle_nested_data_keys(data, nested_data, nested_key):
    if isinstance(nested_data, dict):
        for key, value in nested_data.items():
            data["%s_%s" % (nested_key, key)] = value
    elif isinstance(nested_data, list):
        for index, item in enumerate(nested_data):
            for key, value in item.items():
                data["%s_%s_%s" % (nested_key, index, key)] = value



from cliff.app import App


class BaseApp(App):
    """Overwrite ``cliff.app.App`` class.

    Make possibility to further extension and overwrite methods.
    """
    middlewares = []

    def prepare_to_run_command(self, cmd):
        for m in self.middlewares:
            m.before(cmd)

    def clean_up(self, cmd, result, err):
        for m in reversed(self.middlewares):
            m.after(cmd, result, err)



"""
    DigiCloud Volume Service.
"""
from rich.prompt import Confirm

from .base import Lister, ShowOne, Command
from digicloud import schemas
from ..error_handlers import CLIError
from ..utils import is_tty


def prepare_volume_details_output(session, volume_details):
    # TODO: Refactor required ( cry for help level)
    if volume_details.get('attached_to'):
        attachment_info = volume_details['attached_to'][0]
        volume_details['attachment_info'] = volume_details['attached_to'][0]
        data = session.get('/instances/%s' % attachment_info['instance_id'])
        attachment_info['instance'] = data['name']
    else:
        attachment_info = {}
    volume_details['attachment_info'] = attachment_info
    return volume_details


class ListVolume(Lister):
    """List volumes."""
    schema = schemas.VolumeList(many=True)

    def get_data(self, parsed_args):
        volumes = self.app.session.get('/volumes')
        for volume in volumes:
            volume_instances = []
            for attachment in volume.get('attached_to', []):
                instance_name = attachment['instance_id']
                volume_instances.append(instance_name)
            volume["attached_to"] = ', '.join(volume_instances)
        return volumes


class ShowVolume(ShowOne):
    """Show volume details."""
    schema = schemas.VolumeDetails()

    def get_parser(self, prog_name):
        parser = super(ShowVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help='Volume name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/volumes/%s' % parsed_args.volume
        data = self.app.session.get(uri)
        data = prepare_volume_details_output(self.app.session, data)
        return data


class DeleteVolume(Command):
    """Delete volume."""

    def get_parser(self, prog_name):
        parser = super(DeleteVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help='Volume name or ID'
        )
        parser.add_argument(
            '--i-am-sure',
            help='Use this switch to bypass confirmation',
            default=None,
            action='store_true'
        )
        return parser

    def take_action(self, parsed_args):
        if not self.confirm(parsed_args):
            return
        uri = '/volumes/%s' % parsed_args.volume
        self.app.session.delete(uri)

    def confirm(self, parsed_args):
        if parsed_args.i_am_sure:
            return True
        if is_tty():
            volume = self.app.session.get('/volumes/%s' % parsed_args.volume)
            user_response = Confirm.ask(
                "You're about to delete a volume named "
                "[red bold]{} with {}GB size[/red bold]. "
                "Are you sure?".format(
                    volume['name'],
                    volume['size']
                ))
            if user_response:
                return True
            self.app.stdout.write("Operation cancelled by user\n")
        else:
            self.app.stderr.write("Unable to perform 'delete volume' operation in non-interactive mode,"
                                  " without '--i-am-sure' switch\n")
            return False


class CreateVolume(ShowOne):
    """Create Volume."""
    schema = schemas.VolumeDetails()

    def get_parser(self, prog_name):
        parser = super(CreateVolume, self).get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<name>',
            help='Volume name'
        )
        parser.add_argument(
            '--size',
            required=True,
            metavar='<size>',
            help='Volume size'
        )

        parser.add_argument(
            '--type',
            required=True,
            metavar='<type>',
            help='Volume type, could be SSD or ULTRA_DISK',
            choices=['SSD', 'ULTRA_DISK'],
            type=lambda value: str(value).upper().replace("-", "_"),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help='Volume description'
        )

        return parser

    def get_data(self, parsed_args):
        volume_type = parsed_args.type.upper()
        payload = {
            'name': parsed_args.name,
            'size': parsed_args.size,
            'volume_type': volume_type
        }
        if parsed_args.description:
            payload['description'] = parsed_args.description

        data = self.app.session.post('/volumes', payload)
        data = prepare_volume_details_output(self.app.session, data)

        return data


class UpdateVolume(ShowOne):
    """Update volume."""
    schema = schemas.VolumeDetails()

    def get_parser(self, prog_name):
        parser = super(UpdateVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help=('Volume name or ID'),
        )
        parser.add_argument(
            '--name',
            required=False,
            metavar='<name>',
            help='new name for the volume, must be unique',
        )
        parser.add_argument(
            '--description',
            required=False,
            metavar='<description>',
            help='Volume description'
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/volumes/%s' % parsed_args.volume
        payload = {}
        if parsed_args.name:
            payload['name'] = parsed_args.name
        if parsed_args.description:
            payload['description'] = parsed_args.description
        if not payload:
            raise CLIError([dict(
                msg="At least one of --name or --description is necessary"
            )])
        data = self.app.session.patch(uri, payload)
        data = prepare_volume_details_output(self.app.session, data)

        return data


class AttachVolume(ShowOne):
    """Attach instance volume."""
    schema = schemas.InstanceVolume()

    def get_parser(self, prog_name):
        parser = super(AttachVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help='Volume name or ID'
        )
        parser.add_argument(
            '--instance',
            required=True,
            metavar='<instance>',
            help='Instance name or ID'
        )

        return parser

    def get_data(self, parsed_args):
        uri = '/instances/%s/volumes' % parsed_args.instance
        payload = {'id': parsed_args.volume}

        data = self.app.session.post(uri, payload)

        return data


class DetachVolume(Command):
    """Detach instance volume."""

    def get_parser(self, prog_name):
        parser = super(DetachVolume, self).get_parser(prog_name)
        parser.add_argument(
            'volume',
            metavar='<volume>',
            help='Volume name or ID'
        )
        parser.add_argument(
            '--instance',
            required=True,
            metavar='<instance>',
            help='Instance name or ID'
        )

        return parser

    def take_action(self, parsed_args):
        uri = '/instances/%s/volumes/%s' % (parsed_args.instance, parsed_args.volume)
        self.app.session.delete(uri)



import logging
import signal
import sys
import time
import uuid
from urllib.parse import urlparse

from rich.console import Console

from digicloud.sentry import initialize_sentry
from digicloud.utils import get_latest_version_number, compare_versions, is_tty
from digicloud.managers import ConfigManager, Session

logger = logging.getLogger(__name__)

__all__ = [
    'ConfigMiddleware',
    'SignalHandlerMiddleware',
    'VersionCheckMiddleware',
    'SessionMiddleware',
    'RichMiddleware',
    'SentryMiddleware',
]


class BaseMiddleware:
    def __init__(self, app):
        self.app = app

    def before(self, cmd):
        """
        This hook provide control to access the app and the command object
        before command execution
        """

    def after(self, cmd, result, err):
        """
        This hook provide control to access the app, command, result,
        and error if available after command execution
        """


class SignalHandlerMiddleware(BaseMiddleware):
    def before(self, cmd):
        signal.signal(signal.SIGINT, self.interrupt_handler)
        logger.debug("Signal handler registered")

    def interrupt_handler(self, sig, frame):
        print('\nOperation canceled!!')
        sys.exit(130)


class ConfigMiddleware(BaseMiddleware):
    def __init__(self, app):
        self.config = None
        super(ConfigMiddleware, self).__init__(app)

    def before(self, cmd):
        self.config = self.app.config = ConfigManager(self.app.LOG)
        self.config['options'] = {}

    def after(self, cmd, result, err):
        self.config.save()
        logger.debug("Configuration saved")


class VersionCheckMiddleware(BaseMiddleware):

    def __init__(self, app):
        self.session = None
        super(VersionCheckMiddleware, self).__init__(app)

    def before(self, cmd):
        self._check_version()

    def _set_last_run(self):
        self.app.config['last_version_check_time'] = int(time.time())

    def _get_last_run(self):
        now = int(time.time())
        if 'last_version_check_time' in self.app.config:
            return now - int(self.app.config['last_version_check_time'])
        return 0

    def _check_version(self):
        try:
            latest_version = get_latest_version_number('digicloud')
            current_version = self.app.current_version
            diff = compare_versions(latest_version, current_version)
            if diff['major']:
                self._important_version_change(current_version, latest_version)
            if diff['minor'] or diff['patch']:
                self._regular_version_change(current_version, latest_version)
        except Exception as exp:
            if getattr(self.app.options, "debug", False):
                raise exp

    def _important_version_change(self, current_version, latest_version):
        self.app.console.print(
            "[red bold]You current version is {} which is no longer "
            "work properly due to recent changes.\n[/red bold]"
            "Please update to latest version via "
            "[blue bold]pip install digicloud -U[blue bold]".format(
                current_version
            )
        )
        sys.exit(1)

    def _regular_version_change(self, current_version, latest_version):
        last_run = self._get_last_run()
        should_warn = (last_run == 0 or last_run > 3600) and is_tty()
        if should_warn:
            self.app.console.print(
                "[red bold]A new Digicloud "
                "version ({}) is available.[red bold]".format(
                    latest_version
                )
            )
        self._set_last_run()


class SessionMiddleware(BaseMiddleware):
    def __init__(self, app):
        self.session = getattr(app, 'session', None)
        super(SessionMiddleware, self).__init__(app)

    def before(self, cmd):
        config = self.app.config
        base_url = config['BASE_URL']
        session = config.get('SESSION', Session(base_url))
        session.user_agent = "PYTHON-CLI_{}".format(self.app.current_version)
        self.app.session = self.session = session
        session.setup(config)

    def after(self, cmd, result, err):
        self.app.config['SESSION'] = self.session


class RichMiddleware(BaseMiddleware):
    def before(self, cmd):
        self.app.console = Console()


class SentryMiddleware(BaseMiddleware):

    def before(self, cmd):
        if "client_id" not in self.app.config:
            self.app.console.print(
                "[yellow bold]"
                "Digicloud will collect errors and related context anonymously.\n"
                "[/yellow bold]"
                "To disable it you can easily run: "
                "[blue bold]digicloud config --set collect_error_info=0[/blue bold]"
                ""
            )
            self.app.config['client_id'] = uuid.uuid4().hex
        client_id = self.app.config['client_id']
        base_url = self.app.config['BASE_URL']
        region = self.app.config['AUTH_HEADERS'].get('Digicloud-Region', "")
        namespace = self.app.config['AUTH_HEADERS'].get('Digicloud-Namespace', "")
        initialize_sentry(
            self.app.current_version,
            urlparse(base_url).netloc,
            region=region,
            namespace=namespace,
            client_id=client_id
        )



import os

from sentry_sdk import init, set_tag


def strip_sensitive_data(event, hint):
    if 'password' in str(event).lower():
        # skip event containing sensitive information
        return None
    if 'log_record' in hint:
        # Skip the logs
        return None
    return event


def initialize_sentry(app_version, server_name, **tags):
    init(
        os.getenv(
            'SENTRY_URL',
            'https://19915a046a814dfc9940040b4f8f83b3@yggdntry.digicloud.ir//2'
        ),
        traces_sample_rate=1.0,
        before_send=strip_sensitive_data,
        release=app_version,
        server_name=server_name,
        shutdown_timeout=1
    )
    for key, value in tags.items():
        set_tag(key, value)



"""
    Configure ``setup.py`` for digicloud cli app.
"""
from setuptools import setup, find_packages


try:
    LONG_DESCRIPTION = open('README.md', 'rt').read()
except IOError:
    LONG_DESCRIPTION = ''

setup(
    name='digicloud',
    version='2.5.17',
    description='DigiCloud CLI',

    long_description=LONG_DESCRIPTION,

    author='Arash Shams',
    author_email='a.shams@digikala.com',
    url='https://digicloud.ir',
    download_url='',
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Intended Audience :: Developers',
        'Environment :: Console',
    ],
    platforms=['Any'],
    scripts=[],
    provides=[],
    install_requires=[
        'cliff',
        'rich',
        'requests',
        'marshmallow',
        'setuptools',
        'tzlocal',
        'pytz',
        'sentry-sdk',
        'tabulate',
    ],
    extras_require={
        'dev': [
            'pytest==6.1.0'
        ]
    },
    packages=find_packages(exclude=["tests"]),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'digicloud = digicloud.main:main'
        ],
        'digicloud.formatters.list': [
            'rich = digicloud.cli.formatters:RichFormatter',
            'table = cliff.formatters.table:TableFormatter',
            'json = cliff.formatters.json_format:JSONFormatter',
            'yaml = cliff.formatters.yaml_format:YAMLFormatter',
            'value = cliff.formatters.value:ValueFormatter',
        ],
        'digicloud.formatters.details': [
            'rich = digicloud.cli.formatters:RichFormatter',
            'table = cliff.formatters.table:TableFormatter',
            'json = cliff.formatters.json_format:JSONFormatter',
            'yaml = cliff.formatters.yaml_format:YAMLFormatter',
            'value = cliff.formatters.value:ValueFormatter',
        ],
        'digicloud.cli': [
            'explain = digicloud.commands.explain:Explain',
            'config = digicloud.commands.config:Config',
            'account_login = digicloud.commands.account:Login',
            'account_logout = digicloud.commands.account:Logout',
            'account_current = digicloud.commands.account:CurrentUser',
            'namespace_current = digicloud.commands.namespace:CurrentNamespace',
            'namespace_select = digicloud.commands.namespace:SelectNamespace',
            'namespace_create = digicloud.commands.namespace:CreateNamespace',
            'namespace_leave = digicloud.commands.namespace:LeaveNamespace',
            'namespace_list = digicloud.commands.namespace:ListNamespace',
            'namespace_show = digicloud.commands.namespace:ShowNamespace',
            'namespace_member_list = digicloud.commands.namespace:ListNamespaceMember',
            'namespace_member_add = digicloud.commands.namespace:InviteMember',
            'namespace_invitation_accept = digicloud.commands.namespace:AcceptInvitation',
            'namespace_member_delete = digicloud.commands.namespace:DeleteNamespaceMember',
            'namespace_quota_list = digicloud.commands.namespace:ListQuota',
            'namespace_quota_request = digicloud.commands.quota:RequestMoreQuota',
            'namespace_quota_request_list = digicloud.commands.quota:QuotaRequestList',
            'namespace_invitation_list = digicloud.commands.namespace:ListUserInvitations',
            'instance_type_list = digicloud.commands.instance_type:ListInstanceType',
            'instance_type_show = digicloud.commands.instance_type:ShowInstanceType',
            'public_ip_list = digicloud.commands.access:ListPublicIp',
            'public_ip_show = digicloud.commands.access:ShowPublicIp',
            'public_ip_create = digicloud.commands.access:CreatePublicIp',
            'public_ip_delete = digicloud.commands.access:DeletePublicIp',
            'public_ip_associate = digicloud.commands.access:AssociatePublicIp',
            'public_ip_revoke = digicloud.commands.access:RevokePublicIp',
            'image_list = digicloud.commands.image:ListImage',
            'image_show = digicloud.commands.image:ShowImage',
            'instance_list = digicloud.commands.instance:ListInstance',
            'instance_create = digicloud.commands.instance:CreateInstance',
            'instance_delete = digicloud.commands.instance:DeleteInstance',
            'instance_update = digicloud.commands.instance:UpdateInstance',
            'instance_resize = digicloud.commands.instance:ResizeInstance',
            'instance_start = digicloud.commands.instance_operations:StartInstance',
            'instance_stop = digicloud.commands.instance_operations:StopInstance',
            'instance_suspend = digicloud.commands.instance_operations:SuspendInstance',
            'instance_resume = digicloud.commands.instance_operations:ResumeInstance',
            'instance_reboot = digicloud.commands.instance_operations:RebootInstance',
            'instance_show = digicloud.commands.instance:ShowInstance',
            'instance_volume_list = digicloud.commands.instance:ListInstanceVolume',
            'instance_volume_show = digicloud.commands.instance:ShowInstanceVolume',
            'instance_volume_attach = digicloud.commands.instance:AttachInstanceVolume',
            'instance_volume_detach = digicloud.commands.instance:DetachInstanceVolume',
            'instance_interface_list = digicloud.commands.instance:ListInstanceInterface',
            'instance_interface_attach = digicloud.commands.instance:AttachInterface',
            'instance_interface_detach = digicloud.commands.instance:DetachInterface',
            'instance_firewall_list = digicloud.commands.instance:ListFirewall',
            'instance_firewall_add = digicloud.commands.instance:AddFirewall',
            'instance_firewall_remove = digicloud.commands.instance:RemoveFirewall',
            'instance_snapshot_list = digicloud.commands.instance:ListInstanceSnapshot',
            'instance_log_show = digicloud.commands.instance:InstanceConsoleLog',
            'snapshot_create = digicloud.commands.snapshot:CreateSnapshot',
            'snapshot_list = digicloud.commands.snapshot:ListSnapshot',
            'snapshot_show = digicloud.commands.snapshot:ShowSnapshot',
            'snapshot_delete = digicloud.commands.snapshot:DeleteSnapshot',
            'snapshot_update = digicloud.commands.snapshot:UpdateSnapshot',
            'volume_list = digicloud.commands.volume:ListVolume',
            'volume_show = digicloud.commands.volume:ShowVolume',
            'volume_delete = digicloud.commands.volume:DeleteVolume',
            'volume_create = digicloud.commands.volume:CreateVolume',
            'volume_update = digicloud.commands.volume:UpdateVolume',
            'volume_attach = digicloud.commands.volume:AttachVolume',
            'volume_detach = digicloud.commands.volume:DetachVolume',
            'network_list = digicloud.commands.network:ListNetwork',
            'network_show = digicloud.commands.network:ShowNetwork',
            'network_delete = digicloud.commands.network:DeleteNetwork',
            'network_create = digicloud.commands.network:CreateNetwork',
            'network_update = digicloud.commands.network:UpdateNetwork',
            'network_port_list = digicloud.commands.network:ListNetworkPorts',
            'firewall_list = digicloud.commands.security:ListFirewall',
            'firewall_show = digicloud.commands.security:ShowFirewall',
            'firewall_create = digicloud.commands.security:CreateFirewall',
            'firewall_delete = digicloud.commands.security:DeleteFirewall',
            'firewall_update = digicloud.commands.security:UpdateFirewall',
            'firewall_rule_list = digicloud.commands.security:ListFirewallRule',
            'firewall_rule_show = digicloud.commands.security:ShowFirewallRule',
            'firewall_rule_create = digicloud.commands.security:CreateFirewallRule',
            'firewall_rule_delete = digicloud.commands.security:DeleteFirewallRule',
            'subnet_create = digicloud.commands.subnet:CreateSubnet',
            'subnet_list = digicloud.commands.subnet:ListSubnet',
            'subnet_show = digicloud.commands.subnet:ShowSubnet',
            'subnet_delete = digicloud.commands.subnet:DeleteSubnet',
            'subnet_update = digicloud.commands.subnet:UpdateSubnet',
            'ssh_key_list = digicloud.commands.ssh_key:ListSSHKey',
            'ssh_key_show = digicloud.commands.ssh_key:ShowSSHKey',
            'ssh_key_delete = digicloud.commands.ssh_key:DeleteSSHKey',
            'ssh_key_create = digicloud.commands.ssh_key:CreateSSHKey',
            'router_list = digicloud.commands.router:ListRouter',
            'router_show = digicloud.commands.router:ShowRouter',
            'router_update = digicloud.commands.router:UpdateRouter',
            'router_delete = digicloud.commands.router:DeleteRouter',
            'router_create = digicloud.commands.router:CreateRouter',
            'router_interface_list = digicloud.commands.router:ListRouterInterface',
            'router_interface_show = digicloud.commands.router:ShowRouterInterface',
            'router_interface_add = digicloud.commands.router:AddRouterInterface',
            'router_interface_remove = digicloud.commands.router:RemoveRouterInterface',
            'router_external_add = digicloud.commands.router:AddRouterExternal',
            'router_external_remove = digicloud.commands.router:RemoveRouterExternal',
            'router_static_list = digicloud.commands.router:ListRouterStatic',
            'router_static_add = digicloud.commands.router:AddRouterStatic',
            'router_static_delete = digicloud.commands.router:DeleteRouterStatic',
            'region_list = digicloud.commands.region:ListRegions',
            'region_select = digicloud.commands.region:SelectRegion',
            'region_current = digicloud.commands.region:CurrentRegion',
            'vpn_external_create = digicloud.commands.vpn:CreateExternalVpnConnection',
            'vpn_external_list = digicloud.commands.vpn:ListExternalVpnConnections',
            'vpn_external_delete = digicloud.commands.vpn:DeleteExternalVpnConnection',
            'vpn_external_update = digicloud.commands.vpn:UpdateExternalVpnConnection',
            'vpn_external_show = digicloud.commands.vpn:ShowExternalVpnConnections',

            'edge_domain_list = digicloud.commands.edge:ListDomain',
            'edge_domain_create = digicloud.commands.edge:CreateDomain',
            'edge_domain_delete = digicloud.commands.edge:DeleteDomain',
            'edge_domain_show = digicloud.commands.edge:ShowDomain',
            'edge_record_list = digicloud.commands.edge:ListRecord',
            'edge_record_show = digicloud.commands.edge:ShowRecord',
            'edge_record_create = digicloud.commands.edge:CreateRecord',
            'edge_record_delete = digicloud.commands.edge:DeleteRecord',
            'edge_record_update = digicloud.commands.edge:UpdateRecord',
            'edge_upstream_list = digicloud.commands.edge:ListUpstream',
            'edge_upstream_show = digicloud.commands.edge:ShowUpstream',
            'edge_upstream_create = digicloud.commands.edge:CreateUpstream',
            'edge_upstream_delete = digicloud.commands.edge:DeleteUpstream',
            'edge_upstream_update = digicloud.commands.edge:UpdateUpstream',
            'edge_upstream_server_list = digicloud.commands.edge:ListUpstreamServer',
            'edge_upstream_server_show = digicloud.commands.edge:ShowUpstreamServer',
            'edge_upstream_server_create = digicloud.commands.edge:CreateUpstreamServer',
            'edge_upstream_server_delete = digicloud.commands.edge:DeleteUpstreamServer',
            'edge_upstream_server_update = digicloud.commands.edge:UpdateUpstreamServer',
            'edge_ssl_show = digicloud.commands.edge:ShowSSL',
            'edge_ssl_update = digicloud.commands.edge:UpdateSSL',
            'edge_location_list = digicloud.commands.edge:ListLocation',
            'edge_location_show = digicloud.commands.edge:ShowLocation',
            'edge_location_create = digicloud.commands.edge:CreateLocation',
            'edge_location_delete = digicloud.commands.edge:DeleteLocation',
            'edge_location_update = digicloud.commands.edge:UpdateLocation',
            'edge_firewall_list = digicloud.commands.edge:ListFirewall',
            'edge_firewall_show = digicloud.commands.edge:ShowFirewall',
            'edge_firewall_create = digicloud.commands.edge:CreateFirewall',
            'edge_firewall_delete = digicloud.commands.edge:DeleteFirewall',
            'edge_firewall_update = digicloud.commands.edge:UpdateFirewall',
            'edge_cache_zone_list = digicloud.commands.edge:ListCacheZone',
            'edge_general_setting_show = digicloud.commands.edge:ShowGeneralSetting',
            'edge_general_setting_update = digicloud.commands.edge:UpdateGeneralSetting',
            'edge_purge_cache = digicloud.commands.edge:PurgeCache',
            'edge_domain_dnssec_show = digicloud.commands.edge:GetDomainDNSSEC',
            'edge_domain_dnssec_enable = digicloud.commands.edge:EnableDomainDNSSEC',
            'edge_domain_dnssec_disable = digicloud.commands.edge:DisableDomainDNSSEC',
            'edge_domain_verify_show = digicloud.commands.edge:GetDomainNSRecordSetting',
            'edge_domain_verify = digicloud.commands.edge:VerifyNSRecord',
            'edge_rate_limit_list = digicloud.commands.edge:ListRateLimit',
            'edge_rate_limit_show = digicloud.commands.edge:ShowRateLimit',
            'edge_rate_limit_create = digicloud.commands.edge:CreateRateLimit',
            'edge_rate_limit_delete = digicloud.commands.edge:DeleteRateLimit',
            'edge_rate_limit_update = digicloud.commands.edge:UpdateRateLimit',
            'loadbalancer_create = digicloud.commands.load_balancer:CreateLoadBalancer',
            'loadbalancer_list = digicloud.commands.load_balancer:ListLoadBalancer',
            'loadbalancer_delete = digicloud.commands.load_balancer:DeleteLoadBalancer',
            'loadbalancer_update = digicloud.commands.load_balancer:UpdateLoadBalancer',
            'loadbalancer_show = digicloud.commands.load_balancer:ShowLoadBalancer',
            'loadbalancer_backend_create = digicloud.commands.load_balancer:CreateBackend',
            'loadbalancer_backend_delete = digicloud.commands.load_balancer:DeleteBackend',
            'loadbalancer_backend_list = digicloud.commands.load_balancer:ListBackend',
            'loadbalancer_frontend_create = digicloud.commands.load_balancer:CreateFrontend',
            'loadbalancer_frontend_set_backend = digicloud.commands.load_balancer:SetFrontendDefaultBackend',
            'loadbalancer_frontend_delete_backend = digicloud.commands.load_balancer:DeleteFrontendDefaultBackend',
            'loadbalancer_frontend_delete = digicloud.commands.load_balancer:DeleteFrontend',
            'loadbalancer_frontend_list = digicloud.commands.load_balancer:ListFrontend',
            'loadbalancer_backend_member_add = digicloud.commands.load_balancer:AddBackendMember',
            'loadbalancer_backend_member_list = digicloud.commands.load_balancer:ListBackendMember',
            'loadbalancer_backend_member_remove = digicloud.commands.load_balancer:DeleteBackendMember',
            'loadbalancer_backend_health_check_list = digicloud.commands.load_balancer:ListHealthCheck',
            'loadbalancer_backend_health_check_create = digicloud.commands.load_balancer:CreateHealthCheck',
            'loadbalancer_backend_health_check_delete = digicloud.commands.load_balancer:DeleteHealthCheck',
        ],
    },
    zip_safe=False,
)



"""
    DigiCloud SubnetList Service.
"""


from ..cli import parseractions
from digicloud import schemas
from .base import Lister, ShowOne, Command
from ..error_handlers import CLIError


class CreateSubnet(ShowOne):
    """Create Network SubnetList"""
    schema = schemas.SubnetDetails()

    def get_parser(self, prog_name):
        parser = super(CreateSubnet, self).get_parser(prog_name)
        subnet_group = parser.add_mutually_exclusive_group()
        parser.add_argument(
            'name',
            metavar='<name>',
            help='SubnetList name'
        )
        parser.add_argument(
            '--ip-version',
            default='4',
            metavar='<ip_version>',
            help='IP version, default to IPv4'
        )
        parser.add_argument(
            '--cidr',
            required=True,
            metavar='<cidr>',
            help='CIDR, e.g: 10.0.0.0/16'
        )
        parser.add_argument(
            '--network',
            required=True,
            metavar='<network>',
            help='network name or ID'
        )
        parser.add_argument(
            '--enable-dhcp',
            metavar='<enable_dhcp>',
            help='Enable DHCP for this subnet'
        )
        subnet_group.add_argument(
            '--disable-gateway',
            action='store_true',
            help='Disable gateway for this subnet'
        )
        subnet_group.add_argument(
            '--gateway-ip',
            metavar='<gateway_ip>',
            help='Set gateway IP for this subnet'
        )
        parser.add_argument(
            '--allocation-pool',
            metavar='start=<ip-address>,end=<ip-address>',
            dest='allocation_pools',
            action=parseractions.MultiKeyValueAction,
            required_keys=['start', 'end'],
            help='Allocation pool IP addresses for this subnet e.g.: start=192.168.199.2,'
                 'end=192.168.199.254 (repeat option to add multiple IP addresses)'
        )
        parser.add_argument(
            '--dns-server',
            metavar='<dns_server>',
            action='append',
            dest='dns_server',
            help="DNS server for this subnet (repeat option to set multiple DNS servers)"
        )
        parser.add_argument(
            '--host-route',
            metavar='destination=<subnet>,nexthop=<ip-address>',
            dest='host_route',
            action=parseractions.MultiKeyValueAction,
            required_keys=['destination', 'nexthop'],
            help="Additional route for this subnet "
                 "e.g.: destination=10.10.0.0/16,nexthop=192.168.71.254 "
                 "destination: destination subnet (in CIDR notation) "
                 "nexthop: nexthop IP address "
                 "(repeat option to add multiple routes)")

        return parser

    def get_data(self, parsed_args):
        uri = '/networks/%s/subnets' % parsed_args.network
        payload = {
            'ip_version': parsed_args.ip_version,
            'cidr': parsed_args.cidr,
            'subnet_name': parsed_args.name,
            **self._get_optional_fields(parsed_args)
        }
        data = self.app.session.post(uri, payload)

        return data

    def _get_optional_fields(self, args):
        optional_args = {}

        if args.enable_dhcp is not None:
            optional_args["enable_dhcp"] = args.enable_dhcp

        # NOTE: It's boolean
        if args.disable_gateway:
            optional_args["gateway_ip"] = None

        if args.gateway_ip is not None:
            optional_args["gateway_ip"] = args.gateway_ip

        if args.allocation_pools is not None:
            optional_args["allocation_pools"] = args.allocation_pools

        if args.dns_server is not None:
            optional_args["dns_servers"] = args.dns_server

        if args.host_route is not None:
            optional_args["host_routes"] = args.host_route

        return optional_args


class ListSubnet(Lister):
    """List subnets."""
    schema = schemas.SubnetList(many=True)

    def get_data(self, parsed_args):
        data = self.app.session.get('/subnets')

        return data


class ShowSubnet(ShowOne):
    """Show subnet details."""
    schema = schemas.SubnetDetails()

    def get_parser(self, prog_name):
        parser = super(ShowSubnet, self).get_parser(prog_name)
        parser.add_argument(
            'subnet',
            metavar='<subnet>',
            help='SubnetList name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/subnets/%s' % parsed_args.subnet
        data = self.app.session.get(uri)

        return data


class DeleteSubnet(Command):
    """Delete subnet."""

    def get_parser(self, prog_name):
        parser = super(DeleteSubnet, self).get_parser(prog_name)
        parser.add_argument(
            'subnet',
            metavar='<subnet>',
            help='subnet name or ID'
        )
        return parser

    def take_action(self, parsed_args):
        uri = '/subnets/%s' % parsed_args.subnet
        self.app.session.delete(uri)


class UpdateSubnet(ShowOne):
    """Update subnets."""
    schema = schemas.SubnetDetails()

    def get_parser(self, prog_name):
        parser = super(UpdateSubnet, self).get_parser(prog_name)
        subnet_group = parser.add_mutually_exclusive_group()
        parser.add_argument(
            'subnet',
            metavar='<subnet>',
            help='SubnetList ID',
        )
        parser.add_argument(
            '--name',
            metavar='<Name>',
            help='New name for subnet.'
        )
        parser.add_argument(
            '--description',
            metavar='<Description>',
            help='New description for subnet.'
        )
        parser.add_argument(
            '--dns-server',
            metavar='<dns_server>',
            action='append',
            dest='dns_server',
            help="DNS server for this subnet "
                 "(repeat option to set multiple DNS servers - old nameservers will be deleted)"
        )
        subnet_group.add_argument(
            '--disable-gateway',
            action='store_true',
            help='Disable gateway for this subnet'
        )
        subnet_group.add_argument(
            '--gateway-ip',
            metavar='<gateway_ip>',
            help='Set gateway IP for this subnet'
        )
        parser.add_argument(
            '--allocation-pool',
            metavar='start=<ip-address>,end=<ip-address>',
            dest='allocation_pools',
            action=parseractions.MultiKeyValueAction,
            required_keys=['start', 'end'],
            help='Allocation pool IP addresses for this subnet e.g.: start=192.168.199.2,'
                 'end=192.168.199.254 (repeat option to add multiple IP addresses)'
        )
        parser.add_argument(
            '--host-route',
            metavar='destination=<subnet>,nexthop=<ip-address>',
            dest='host_route',
            action=parseractions.MultiKeyValueAction,
            required_keys=['destination', 'nexthop'],
            help="Additional route for this subnet "
                 "e.g.: destination=10.10.0.0/16,nexthop=192.168.71.254 "
                 "destination: destination subnet (in CIDR notation) "
                 "nexthop: nexthop IP address "
                 "(repeat option to add multiple routes)"
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/subnets/%s' % parsed_args.subnet
        payload = {}
        if parsed_args.name:
            payload['name'] = parsed_args.name
        if parsed_args.description:
            payload['description'] = parsed_args.description
        if parsed_args.dns_server:
            payload['dns_servers'] = parsed_args.dns_server

        if parsed_args.disable_gateway:
            payload["gateway_ip"] = None

        if parsed_args.gateway_ip is not None:
            payload["gateway_ip"] = parsed_args.gateway_ip

        if parsed_args.allocation_pools is not None:
            payload["allocation_pools"] = parsed_args.allocation_pools

        if parsed_args.host_route is not None:
            payload["host_routes"] = parsed_args.host_route


        if not payload:
            raise CLIError([dict(
                msg="At least one of --name or"
                    " --description  or "
                    " --dns_server or "
                    " --gateway_ip or "
                    " --gateway_ip or "
                    " --allocation_pools or "
                    " --host_routes is necessary"
            )])
        data = self.app.session.patch(uri, payload)
        return data



"""
    DigiCloud S2S VPN service.
"""
from requests import HTTPError

from .base import Lister, ShowOne, Command
from .. import schemas
from ..error_handlers import CLIError


class CreateExternalVpnConnection(ShowOne):
    """Create external IpSec VPN connection."""

    schema = schemas.ExternalVpnConnectionDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'name',
            metavar='<VpnConnectionName>',
            help="The external VPN connection name to be created.",
        )
        parser.add_argument(
            '--description',
            metavar='<VpnConnectionDescription>',
            help="The external VPN connection description.",
        )
        parser.add_argument(
            '--local-endpoint-group',
            metavar='<LocalSubnet>',
            help='List of local subnet names or ids. E.g: subnet_1 subnet_2',
            nargs='+',
            required=True
        )
        parser.add_argument(
            '--peer-id',
            metavar='<PeerId>',
            help='The peer router identity for authentication. '
                 'A valid value is an IPv4 address, IPv6 address or FQDN. '
                 'Typically, this value matches the --peer-address.',
            required=True
        )
        parser.add_argument(
            '--peer-address',
            metavar='<PeerAddress>',
            help="The peer gateway public IPv4 or IPv6 address or FQDN.",
            required=True
        )
        parser.add_argument(
            '--peer-endpoint-group',
            metavar='<PeerCidr>',
            help="list of peer networks in CIDR format. E.g: cidr_1 cidr_2",
            nargs='+',
            required=True
        )
        parser.add_argument(
            '--psk',
            metavar='<PSK>',
            help="Pre-shared shared key for the VPN connection.",
            required=True
        )
        parser.add_argument(
            '--initiator',
            metavar='<Initiator>',
            help="Indicates whether this VPN can only respond to connections or both respond to "
                 "and initiate connections. "
                 "A valid value is 'response-only' or 'bi-directional'. "
                 "Default is 'bi-directional'.",
            default="bi-directional",
            choices=['response-only', 'bi-directional']
        )
        parser.add_argument(
            '--admin-state-down',
            help="Sets the administrative state of the resource to 'down', "
                 "including this switch means 'down' state. "
                 "omitting this switch means 'up' state",
            default=False,
            action="store_true"
        )
        parser.add_argument(
            '--mtu',
            metavar='<MTU>',
            help="The maximum transmission unit (MTU) value to address fragmentation. "
                 "Minimum value is 68 for IPv4, and 1280 for IPv6.",
            required=True,
            type=int
        )
        parser.add_argument(
            '--ike-auth-algorithm',
            metavar='<IkeAuthAlgorithm>',
            help="The IKE authentication hash algorithm. "
                 "Valid values are 'sha1', 'sha256', 'sha384', 'sha512'. "
                 "The default is 'sha1'.",
            default="sha1",
            choices=['sha1', 'sha256', 'sha384', 'sha512']
        )
        parser.add_argument(
            '--ike-encryption-algorithm',
            metavar='<IkeEncryptionAlgorithm>',
            help="The IKE encryption algorithm. "
                 "A valid value is '3des', 'aes-128', 'aes-192', 'aes-256'. "
                 "Default is 'aes-128'.",
            default="aes-128",
            choices=['3des', 'aes-128', 'aes-192', 'aes-256']
        )
        parser.add_argument(
            '--ike-pfs',
            metavar='<IkePsf>',
            help="Perfect forward secrecy (PFS). A valid value is 'group2', 'group5', 'group14'. "
                 "Default is 'group5'.",
            default="group5",
            choices=['group2', 'group5', 'group14']
        )
        parser.add_argument(
            '--ike-lifetime',
            metavar='<IkeLifetime>',
            help="IKE lifetime in seconds. Default is 3600.",
            default=3600,
            type=int
        )
        parser.add_argument(
            '--ike-version',
            metavar='<IkeVersion>',
            help="The IKE version. A valid value is 'v1' or 'v2'. Default is 'v1'.",
            default="v1",
            choices=['v1', 'v2']
        )
        parser.add_argument(
            '--vpn-router-id',
            metavar='<VpnRouterId>',
            help="The name or id of the router to be used with the VPN",
            required=True
        )
        parser.add_argument(
            '--ipsec-auth-algorithm',
            metavar='<IpsecAuthAlgorithm>',
            help="The IpSec authentication hash algorithm. "
                 "Valid values are 'sha1', 'sha256', 'sha384', 'sha512'. "
                 "The default is 'sha1'.",
            default="sha1",
            choices=['sha1', 'sha256', 'sha384', 'sha512']
        )
        parser.add_argument(
            '--ipsec-encapsulation-mode',
            metavar='<IpsecEncapsulationMode>',
            help="The IpSec encapsulation mode. "
                 "A valid value is 'tunnel' or 'transport'. Default is 'tunnel'.",
            default="tunnel",
            choices=['tunnel', 'transport']
        )
        parser.add_argument(
            '--ipsec-encryption-algorithm',
            metavar='<IpsecEncryptionAlgorithm>',
            help="The IpSec encryption algorithm. "
                 "A valid value is '3des', 'aes-128', 'aes-192', 'aes-256'. "
                 "Default is 'aes-128'.",
            default="aes-128",
            choices=['3des', 'aes-128', 'aes-192', 'aes-256']
        )
        parser.add_argument(
            '--ipsec-pfs',
            metavar='<IpsecPsf>',
            help="The IpSec perfect forward secrecy (PFS)."
                 " A valid value is 'group2', 'group5', 'group14', "
                 "Default is 'group5'.",
            default="group5",
            choices=['group2', 'group5', 'group14']
        )
        parser.add_argument(
            '--ipsec-transform-protocol',
            metavar='<IpsecTransform>',
            help="The IpSec transform protocol. "
                 "A valid value is 'esp', 'ah', or 'ah-esp'. Default is 'esp'.",
            default="esp",
            choices=['esp', 'ah', 'ah-esp']
        )
        parser.add_argument(
            '--ipsec-lifetime',
            metavar='<IpsecLifetime>',
            help="The IpSec lifetime in seconds. Default is 3600.",
            default=3600,
            type=int
        )
        return parser

    def get_data(self, parsed_args):
        local_endpoint_group = self._validate_local_endpoint_group(
            parsed_args.local_endpoint_group
        )
        vpn_router = self._validate_router(parsed_args.vpn_router_id)

        payload = {
            "name": parsed_args.name,
            "local_endpoint_group": {"endpoints": local_endpoint_group},
            "peer_id": parsed_args.peer_id,
            "peer_address": parsed_args.peer_address,
            "peer_endpoint_group": {"endpoints": parsed_args.peer_endpoint_group},
            "psk": parsed_args.psk,
            "initiator": parsed_args.initiator,
            "mtu": parsed_args.mtu,
            "ike_policy": {
                "auth_algorithm": parsed_args.ike_auth_algorithm,
                "encryption_algorithm": parsed_args.ike_encryption_algorithm,
                "pfs": parsed_args.ike_pfs,
                "lifetime": {"value": parsed_args.ike_lifetime},
                "ike_version": parsed_args.ike_version
            },
            "vpn_service": {"router_id": vpn_router['id']},
            "ipsec_policy": {
                "auth_algorithm": parsed_args.ipsec_auth_algorithm,
                "encapsulation_mode": parsed_args.ipsec_encapsulation_mode,
                "encryption_algorithm": parsed_args.ipsec_encryption_algorithm,
                "pfs": parsed_args.ipsec_pfs,
                "transform_protocol": parsed_args.ipsec_transform_protocol,
                "lifetime": {"value": parsed_args.ipsec_lifetime}
            }
        }
        if parsed_args.description is not None:
            payload["description"] = parsed_args.description
        payload["admin_state_up"] = True if not parsed_args.admin_state_down else False
        return self.app.session.post('/external-vpn-connections', payload)

    def _validate_local_endpoint_group(self, local_endpoint_group):
        subnets = self.app.session.get("/subnets")
        mapping = {
            **{s['id']: s for s in subnets},
            **{s['name']: s for s in subnets}
        }
        selected_subnet_ids = []
        for endpoint in local_endpoint_group:
            if endpoint in mapping:
                selected_subnet_ids.append(mapping[endpoint]['id'])
            else:
                raise CLIError([
                    {
                        "msg": "{} is not a valid subnet name or ID".format(
                            endpoint
                        ),
                        "hint": "Use digicloud subnet list to check your existing subnets"
                    }
                ])
        return selected_subnet_ids

    def _validate_router(self, vpn_router_id):
        try:
            return self.app.session.get('/routers/%s' % vpn_router_id)
        except HTTPError as exp:
            if exp.response.status_code == 404:
                raise CLIError([
                    {
                        "msg": "{} is not a valid router name or id".format(
                            vpn_router_id
                        ),
                        "hint": "Use digicloud router list to check your existing routers"
                    }
                ])


class ListExternalVpnConnections(Lister):
    """List external VPN connections."""

    schema = schemas.ExternalVpnConnectionInList(many=True)

    def get_data(self, parsed_args):
        return self.app.session.get('/external-vpn-connections')


class DeleteExternalVpnConnection(Command):
    """Delete external VPN connection."""

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'vpn_connection',
            metavar='<VpnConnection>',
            help='Vpn connection name or Id'
        )
        return parser

    def take_action(self, parsed_args):
        self.app.session.delete('/external-vpn-connections/{}'.format(parsed_args.vpn_connection))


class UpdateExternalVpnConnection(ShowOne):
    """Update external IpSec VPN connection."""
    schema = schemas.ExternalVpnConnectionDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'vpn_connection',
            metavar='<VpnConnection>',
            help="The external VPN connection name or id to be updated.",
        )
        parser.add_argument(
            '--name',
            metavar='<VpnConnectionName>',
            help="The external VPN connection name to be.",
        )
        parser.add_argument(
            '--description',
            metavar='<VpnConnectionDescription>',
            help="The external VPN connection description to be.",
        )
        admin_state_group = parser.add_mutually_exclusive_group()
        admin_state_group.add_argument(
            '--admin-state-down',
            help="Sets the administrative state of the resource to 'down'.",
            action="store_true",
            default=False
        )
        admin_state_group.add_argument(
            '--admin-state-up',
            help="Sets the administrative state of the resource to 'up'.",
            action="store_true",
            default=False
        )
        return parser

    def get_data(self, parsed_args):
        payload = {}
        if parsed_args.name:
            payload["name"] = parsed_args.name
        if parsed_args.description:
            payload["description"] = parsed_args.description

        # these two are mutually exclusive
        if parsed_args.admin_state_up:
            payload["admin_state_up"] = parsed_args.admin_state_up
        if parsed_args.admin_state_down:
            payload["admin_state_up"] = not parsed_args.admin_state_down

        data = self.app.session.patch(
            '/external-vpn-connections/{}'.format(parsed_args.vpn_connection), payload)
        return data


class ShowExternalVpnConnections(ShowOne):
    """Show external VPN connection."""

    schema = schemas.ExternalVpnConnectionDetails()

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'vpn_connection',
            metavar='<VpnConnection>',
            help='Vpn connection name or Id'
        )
        return parser

    def get_data(self, parsed_args):
        return self.app.session.get(
            '/external-vpn-connections/{}'.format(parsed_args.vpn_connection))



"""
    DigiCloud Image Service
"""

from digicloud import schemas
from digicloud.commands.base import Lister, ShowOne


class ListImage(Lister):
    """List images"""
    schema = schemas.ImageList(many=True)

    def get_data(self, parsed_args):
        data = self.app.session.get('/images')
        return data


class ShowImage(ShowOne):
    """Show image details."""
    schema = schemas.ImageDetails()

    def get_parser(self, prog_name):
        parser = super(ShowImage, self).get_parser(prog_name)
        parser.add_argument(
            'image',
            metavar='<image>',
            help='Image name or ID',
        )
        return parser

    def get_data(self, parsed_args):
        uri = '/images/%s' % parsed_args.image
        data = self.app.session.get(uri)
        return data
